@@ -19,7 +19,7 @@ use constant DB_VER_DW        => 0x00030002;
 use constant DB_FLAG_RIJNDAEL => 2;
 use constant DB_FLAG_TWOFISH  => 8;
 
-our $VERSION = '2.00';
+our $VERSION = '2.01';
 my %locker;
 my $salsa20_iv = "\xe8\x30\x09\x4b\x97\x20\x5d\x2a";
 my $qr_date = qr/^(\d\d\d\d)-(\d\d)-(\d\d)[T ](\d\d):(\d\d):(\d\d)(\.\d+|)?Z?$/;
@@ -204,8 +204,8 @@ sub _parse_v1_body {
     $buffer = $self->decrypt_rijndael_cbc($buffer, $key, $head->{'enc_iv'});
 
     die "The file could not be decrypted either because the key is wrong or the file is damaged.\n"
-        if length($buffer) > 2**31 || (!length($buffer) && $head->{'n_groups'});
-    die "The file checksum did not match.\nThe key is wrong or the file is damaged (or we need to implement utf8 input a bit better)\n"
+        if length($buffer) > 2**32-1 || (!length($buffer) && $head->{'n_groups'});
+    die "The file checksum did not match.\nThe key is wrong or the file is damaged\n"
         if $head->{'checksum'} ne sha256($buffer);
 
     my ($groups, $gmap, $pos) = $self->_parse_v1_groups($buffer, $head->{'n_groups'});
@@ -581,7 +581,7 @@ sub _master_key {
             : ($pass && $file) ? sha256($pass, $file) : $pass ? $pass : $file;
     $head->{'enc_iv'}     ||= join '', map {chr rand 256} 1..16;
     $head->{'seed_rand'}  ||= join '', map {chr rand 256} 1..($head->{'version'} && $head->{'version'} eq '2' ? 32 : 16);
-    $head->{'seed_key'}   ||= sha256(time.rand().$$);
+    $head->{'seed_key'}   ||= sha256(time.rand(2**32-1).$$);
     $head->{'rounds'} ||= $self->{'rounds'} || ($head->{'version'} && $head->{'version'} eq '2' ? 6_000 : 50_000);
 
     my $cipher = Crypt::Rijndael->new($head->{'seed_key'}, Crypt::Rijndael::MODE_ECB());
@@ -613,7 +613,7 @@ sub gen_db {
     die "Missing pass\n" if ! defined($pass);
     die "Please unlock before calling gen_db\n" if $self->is_locked($groups);
 
-    srand((time() ^ $$) * rand()) if ! $self->{'no_srand'};
+    srand(rand(time() ^ $$)) if ! $self->{'no_srand'};
     if ($v eq '2') {
         return $self->_gen_v2_db($pass, $head, $groups);
     } else {
@@ -664,7 +664,8 @@ sub _gen_v1_db {
                  [7,      pack('LL', 4, $g->{'icon'}  || 0)],
                  [8,      pack('LS', 2, $g->{'level'} || 0)],
                  [0xFFFF, pack('L', 0)]);
-        push @d, [$_, $g->{'unknown'}->{$_}] for keys %{ $g->{'unknown'} || {} };
+        push @d, [$_, map {pack('L',length $_).$_} $g->{'unknown'}->{$_}]
+            for grep {/^\d+$/ && $_ > 8} keys %{ $g->{'unknown'} || {} };
         $buffer .= pack('S',$_->[0]).$_->[1] for sort {$a->[0] <=> $b->[0]} @d;
         foreach my $e (@{ $g->{'entries'} || [] }) {
             $head->{'n_entries'}++;
@@ -704,7 +705,8 @@ sub _gen_v1_db {
                      [0xD,    pack('L', length($bname)+1)."$bname\0"],
                      [0xE,    pack('L', length($bin)).$bin],
                      [0xFFFF, pack('L', 0)]);
-            push @d, [$_, $e->{'unknown'}->{$_}] for keys %{ $e->{'unknown'} || {} };
+            push @d, [$_, pack('L', length($e->{'unknown'}->{$_})).$e->{'unknown'}->{$_}]
+                for grep {/^\d+$/ && $_ > 0xE} keys %{ $e->{'unknown'} || {} };
             $entries .= pack('S',$_->[0]).$_->[1] for sort {$a->[0] <=> $b->[0]} @d;
         }
     }
@@ -1264,7 +1266,7 @@ sub find_groups {
         $g->{'icon'}  ||= 0;
         while (!defined($g->{'id'}) || $used{$g->{'id'}}++) {
             warn "Found duplicate group_id - generating new one for \"$g->{'title'}\"" if defined($g->{'id'});
-            $g->{'id'} = int((2**32-1) * rand());
+            $g->{'id'} = int(rand 2**32-1);
         }
         if (!@tests || !grep{!$_->($g)} @tests) {
             push @groups, $g;
